/*
  device_ide - VPD support routines for IDE devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_ide_proc.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <linux/hdreg.h>
#include <glob.h>

#include "lsvpd_lib_dir.h"
#include "util.h"
#include "vpd.h"
#include "lsvpd.h"
#include "init.h"

#include "device.h"
#include "device_ide.h"

#include "device_ide_proc.h"

char * device_ide_proc_ide_dir = "/proc/ide";

/********************************************************************/

/*
 * Temporary map created by device_ide_proc_setup (below) and used by
 * other functions.
 */
#define DEVICE_IDE_PROC_MAP "ide_map"

static bool
device_ide_proc_set1(const char * settings,
		     FILE * f)
{
	bool ret;

	char devname[256];
	size_t dl;
	char last;
	char * format;
	unsigned int channel;
	unsigned int unit;

	ret = false;

	format = lsvpd_relative_filename(device_ide_proc_ide_dir,
					 "ide%u/%255[^/]/driver");
	if (NULL != format) {
		if (2 == sscanf(settings, format, &channel, devname)) {
			dl = strlen(devname);
			last = devname[dl - 1];
			unit = 0;
			if (('a' <= last) && (last <= 'z')) 
				unit = ((unsigned int) (last - 'a')) % 2;
			else if (('0' <= last) && (last <= '9'))
				unit = ((unsigned int) (last - '0')) % 2;
			fprintf(f, "%s %d.%d\n", devname, channel, unit);
			ret = true;
		}
		free(format);
	}

	return ret;
}

static bool
device_ide_proc_setup(void)
{
	char * map;
	FILE * f;
	char * pat;
	glob_t globbuf;
	size_t i;

	map = lsvpd_relative_filename(lsvpd_db_misc_dir, DEVICE_IDE_PROC_MAP);
	if (NULL == map)
		goto done;

	f = fopen(map, "w");
	if (NULL == f)
		goto free_map;

	pat = lsvpd_relative_filename(device_ide_proc_ide_dir,
				      "ide*/*/driver");
	if (NULL != pat) {
		if (0 == glob(pat, 0, NULL, &globbuf))
			for (i = 0; i < globbuf.gl_pathc; i++)
				if (!device_ide_proc_set1(globbuf.gl_pathv[i],
							  f))
					break;
		globfree(&globbuf);
		free(pat);
	}
	fclose(f);

 free_map:
	free(map);
 done:
	return true;
}

/********************************************************************/

static bool
device_ide_proc_list_scanner(const char * line,
			     void * x)
{
	char devname[256];
	char cu[256];
	char * t;
	bus_info_list_t ** node_list;

	node_list = x;

	if (2 == sscanf(line, "%255s %255s", devname, cu))
		if (-1 != asprintf(&t, "%s%c%s",
				   DEVICE_TYPE_IDE, BUS_INFO_SEP, cu)) {
			node_list_add(node_list, DEVICE_TYPE_IDE, t);
			free(t);
		}

	return true;
}

static bool
device_ide_proc_list_devices(bus_info_list_t ** node_list)
{
	bool ret;
	char * map;

	ret = false;

	device_ide_proc_setup();

	map = lsvpd_relative_filename(lsvpd_db_misc_dir, DEVICE_IDE_PROC_MAP);
	if (NULL != map) {
		ret = line_scanner_harness(map,
					   device_ide_proc_list_scanner,
					   node_list);
		free(map);
	}

	return ret;
}

/********************************************************************/

typedef struct {
	char * cu;
	char * name;

} proc_ide_get_name_t;

static bool
device_ide_proc_name_scanner(const char * line,
			     void * x)
{
	bool ret;
	char devname[256];
	char cu[256];
	proc_ide_get_name_t * gn;

	gn = x;
	ret = true;

	if ((2 == sscanf(line, "%255s %255s", devname, cu)) &&
	    (0 == strcmp(gn->cu, cu))) {
		gn->name = strdup(devname);
		ret = false;
	}

	return ret;
}	

static char *
device_ide_proc_get_name (const node_t type,
			  const bus_info_t bus_info)
{
	char * ret;
	device_ide_address_t a;
	proc_ide_get_name_t gn;
	char * map;

	ret = NULL;

	gn.name = NULL;

	if ((device_parse_ide_bus_info(bus_info, &a)) &&
	    (-1 != asprintf(&(gn.cu), "%u.%u", a.channel, a.unit))) {
		map = lsvpd_relative_filename(lsvpd_db_misc_dir,
					      DEVICE_IDE_PROC_MAP);
		if (NULL != map) {
			if (line_scanner_harness(map,
						 device_ide_proc_name_scanner,
						 &gn))
				ret = gn.name;
			free(map);
		}
		free(gn.cu);
	}

	return ret;
}

#define DEVICE_IDE_PROC_NUM_DISK_MAJORS 10
static unsigned int
device_ide_proc_disk_majors[DEVICE_IDE_PROC_NUM_DISK_MAJORS] = {
	3, 22, 33,  34,  56,  57, 88,  89,  90, 91,
};

static device_node_info_t *
device_ide_proc_get_info(const node_t type,
			 const bus_info_t bus_info)
{
	device_node_info_t * ret;
	device_ide_address_t a;
	char * name;

	ret = NULL;

	if ((device_parse_ide_bus_info(bus_info, &a)) &&
	    (NULL != (name = node_get_name(type, bus_info)))) {
		if ((0 == strncmp(name, "hd", 2)) &&
		    (DEVICE_IDE_PROC_NUM_DISK_MAJORS > a.channel) &&
		    (NULL != (ret = malloc(sizeof(device_node_info_t))))) {
			ret->major = device_ide_proc_disk_majors[a.channel];
			ret->minor = a.unit ? 64 : 0;
			ret->mode = S_IFBLK;
		} else if ((0 == strncmp(name, "ht", 2)) &&
			   (0 == a.channel) &&
			   (NULL != (ret = malloc(sizeof(device_node_info_t))))) {
			/* devices.txt says only 1 IDE tape drive supported. */
			ret->major = 37;
			ret->minor = 0;
			ret->mode = S_IFCHR;
		}

		free(name);
	}

	return ret;
}

static char *
device_ide_proc_get_subtype(const node_t type,
			    const bus_info_t bus_info)
{
	char * ret;
	char * name;
	char * t;

	ret = NULL;

	if (NULL != (name = node_get_kernel_name(type, bus_info))) {
		if (-1 != asprintf(&t, "%s/%s/media",
				   device_ide_proc_ide_dir, name)) {
			ret = lsvpd_file_read_string(t);
			free(t);
		}
		free(name);
	}

	return ret;
}

#define DEVICE_IDE_PROC_UNKNOWN_DS "IDE Unknown Device"
static struct {
	const char * subtype;
	const char * ds;
} device_ide_proc_ds_map[] = {
	{ "disk",   "IDE Disk Drive"    },
	{ "cdrom",  "IDE CD-ROM Drive"  },
	{ "tape",   "IDE Tape Drive"    },
	{ "floppy", "IDE Floppy Drive"  },
	{ NULL,     DEVICE_IDE_PROC_UNKNOWN_DS },
};

static const char *
device_ide_proc_subtype_to_ds(const char * subtype)
{
	const char * ret;
	int i;
	const char * x;

	ret = DEVICE_IDE_PROC_UNKNOWN_DS;

	if (NULL != subtype) {
		i = 0;
		while ((NULL != (x = device_ide_proc_ds_map[i].subtype)) &&
		       /* Note: strncmp is used here due to laziness.
			* The possible values don't have common
			* prefixes and subtype might have trailing
			* whitespace.  This is easy...
			*/
		       (0 != strncmp(x, subtype, strlen(x))))
			i++;
		ret = device_ide_proc_ds_map[i].ds;
	}

	return ret;
}

static char *
device_ide_proc_get_parent_name(const node_t type,
				const bus_info_t bus_info)
{
	char * ret;
	char * t;
	char * dir;

	ret = NULL;

	t = node_get_kernel_name(type, bus_info);
	if (NULL != t) {
		dir = lsvpd_relative_filename(device_ide_proc_ide_dir, t);
		free(t);
		if (NULL != dir) {
			t = lsvpd_readlink(dir);
			free(dir);
			if (NULL != t) {
				ret = lsvpd_dirname(t);
				free(t);
			}
		}
	}

	return ret;
}

static bool
device_ide_proc_get_identity(const node_t type,
			     const bus_info_t bus_info,
			     struct hd_driveid * id)
{
	bool ret;
	device_node_info_t * i;
	device_temp_t * d;

	ret = false;

	i = device_ide_get_info(type, bus_info);
	if (NULL != i) {
		d = device_open_temporary(i);
		if (NULL != d) {
			ret = (0 == ioctl(d->fd, HDIO_GET_IDENTITY, id));
			device_close_temporary(&d);
		}
		free(i);
	}

	return ret;
}

device_ide_proc_get_identity_t
device_ide_proc_get_identity_function = device_ide_proc_get_identity;

static bool
device_ide_proc_retrieve_vpd(const node_t type,
			     const bus_info_t bus_info,
			     const char * node,
			     vpd_list_t ** vpd)
{
	bool ret;
	char * subtype;
	struct hd_driveid id;
	char * sn;
	char * rm;
	char * tm;
	char * mf;

	ret = false;

	subtype = device_ide_proc_get_subtype(type, bus_info);
	// FIXME: this should not be hardcoded.
	if (!lsvpd_file_relative_write_string(node, LSVPD_SUBTYPE_FILE,
					      NULL == subtype ?
					      "unknown" : subtype))
		goto free_subtype;
		
	if (!vpd_list_maybe_append_key_string(vpd,
					      "DS",
					      device_ide_proc_subtype_to_ds(subtype)))
		goto free_subtype;

	if (!device_ide_proc_get_identity_function(type, bus_info, &id))
		goto free_subtype;

	sn = strndup_trim(id.serial_no, sizeof(id.serial_no));
	rm = strndup_trim(id.fw_rev,    sizeof(id.fw_rev));
	tm = strndup_trim(id.model,     sizeof(id.model));
	device_ide_do_mf_tm(&mf, &tm);

	vpd_list_maybe_append_key_string(vpd, "RM", rm);
	vpd_list_maybe_append_key_string(vpd, "SN", sn);
	vpd_list_maybe_append_key_string(vpd, "TM", tm);
	vpd_list_maybe_append_key_string(vpd, "MF", mf);

	if (NULL != vpd)
		device_ide_vpd_sort(vpd);

	ret = true;

	free(mf);
	free(tm);
	free(sn);
	free(rm);

 free_subtype:
	free(subtype);

	return ret;

}

/********************************************************************/

static void
init(void)
{
	char * f;
	node_functions_t i = {
		.type             = DEVICE_TYPE_IDE,
		.retrieve_vpd     = device_ide_proc_retrieve_vpd,
		.get_parent_name  = device_ide_proc_get_parent_name,
		.get_kernel_name  = device_ide_proc_get_name,
	};

	/* Only register functions if /proc/ide/drivers exists. */
	f = lsvpd_relative_filename(device_ide_proc_ide_dir, "drivers");
	if (NULL != f) {
		if (0 == access(f, F_OK)) {
			node_type_register(&i);
			node_listing_function_register(device_ide_proc_list_devices);
			device_ide_get_info_function = device_ide_proc_get_info;
		}
		free(f);
	}
}

INIT(init);
