/*
  ibm_vpd_render - Convert IBM VPD chunks to text.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: ibm_vpd_render.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "adapter_pci_vpd_parse.h"
#include "growbuf.h"
#include "util.h"
#include "vpd.h"

#define IBM_VPD_SEP_FIELD ((unsigned char *) "FC")
#define IBM_VPD_SEP_VALUE "????????"

static void
usage(const char * prog)
{
	fprintf(stderr, "usage: %s file output-directory\n", prog);
	exit(1);
}

static void
ibm_vpd_add_sep(vpd_list_t ** vl)
{
	vpd_list_prepend_key_value(vl,
				   IBM_VPD_SEP_FIELD,
				   (unsigned char *) IBM_VPD_SEP_VALUE,
				   strlen(IBM_VPD_SEP_VALUE));
}

static vpd_list_list_t *
ibm_vpd_parse(const growbuf * s)
{
	growbuf_iter * i = growbuf_iter_new(s);

	vpd_list_list_t * vll = vpd_list_list_new();

	while (growbuf_iter_have(i, 4)) {
		/* 4 byte, BIG-endian length. */
		uint32_t len = 0;
		int n;
		for (n = 0; n < 4; n++) {
			len = (len << 8) +
				(uint32_t) (* growbuf_iter_current(i));
			growbuf_iter_next(i);
		}
			
		if (growbuf_iter_have(i, len)) {
			vpd_list_t * vl;
			/* Grab PCI VPD data */
			growbuf * pci_vpd =
				growbuf_append(growbuf_new(len),
					       growbuf_iter_current(i),
					       len);
			growbuf_iter_move(i, len);
			
			vl = adapter_pci_vpd_parse(pci_vpd);
			growbuf_free(pci_vpd);
			
			if (NULL != vl) {
				ibm_vpd_add_sep(&vl);
				vll = vpd_list_list_append(vll, vl);
			} else {
				vpd_list_list_free(vll);
				vll = NULL;
				break;
			}
		}
		
	}

	growbuf_iter_free(i);

	return vll;
}

static bool
ibm_vpd_render(const growbuf * vpd, const char * node)
{
	int ret;

	vpd_list_list_t * vll = ibm_vpd_parse(vpd);

	if (NULL == vll)
		return false;

	ret = vpd_list_list_to_node(vll, node);
	vpd_list_list_free(vll);

	return ret;
}

int
main(int argc, char ** argv)
{
	growbuf * vpd;

	unsigned char * data;
	size_t len;

	int ret = 0;

	if (argc != 3)
		usage(argv[0]);

	if (!lsvpd_file_read(argv[1], &data, &len)) {
		fprintf(stderr, "%s: error reading file \"%s\"\n",
			argv[0], argv[1]);
		exit(1);
	}
	vpd = growbuf_append(growbuf_new(len), data, len);
	free(data);

	ret = ibm_vpd_render(vpd, argv[2]) ? 0 : 1;
	growbuf_free(vpd);

	return ret;
}
