/*
  vpd - Utilities for in memory and disk-based VPD handling.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: vpd.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "growbuf.h"
#include "util.h"
#include "vpd.h"

/*
 * VPD_CHECK_NEXT loops until curr is NULL, so VPD_CHECK_NEXT handles
 * the case where we wrap around.
 */
#define VPD_CHECK_NEXT(head, curr) \
  ((NULL != (curr)) && ((curr)->next != head) ? (curr)->next : NULL)

#define VPD_FOREACH(head, curr) \
  for (curr = (head); \
       curr != NULL; \
       curr = VPD_CHECK_NEXT((head), (curr)))

#define VPD_FOREACH_TEMP(head, curr, temp) \
  for (curr = (head), temp = VPD_CHECK_NEXT((head), (curr)); \
       curr != NULL; \
       curr = (temp), temp = VPD_CHECK_NEXT((head), (curr)))

#define VPD_SUBDIR "linux,vpd"
#define VPD_LIST_FILE ".lsvpd"

static vpd_list_t *
vpd_list_new_internal(const vpd_key_t       k,
		      const unsigned char * v,
		      size_t                vlen)
{
	vpd_list_t * vl;

	vl = malloc(sizeof(*vl));
	if (NULL != vl) {
		vl->val = malloc(vlen + 1);
		if (NULL != vl->val) {
			memcpy(vl->val, v, vlen);
			vl->val[vlen] = '\0';
			memcpy(vl->key, k, sizeof(vl->key));
			vl->next = vl;
			vl->prev = vl;
		} else {
			free(vl);
			vl = NULL;
		}
	}

	return vl;
}

vpd_list_t *
vpd_list_new(void)
{
	return NULL;
}

void
vpd_list_free(vpd_list_t * vl)
{
	vpd_list_t * t1;
	vpd_list_t * t2;

	VPD_FOREACH_TEMP(vl, t1, t2) {
		free(t1->val);
		free(t1);
	}
}


static vpd_list_list_t *
vpd_list_list_new_internal(vpd_list_t * vl)
{
	vpd_list_list_t * vll;

	vll = malloc(sizeof(*vll));
	if (NULL != vll) {
		vll->vpd_list = vl;
		vll->prev = vll;
		vll->next = vll;
	}

	return vll;
}

vpd_list_list_t *
vpd_list_list_new(void)
{
	return NULL;
}

void
vpd_list_list_free(vpd_list_list_t * vll)
{
	vpd_list_list_t * t1;
	vpd_list_list_t * t2;

	VPD_FOREACH_TEMP(vll, t1, t2) {
		vpd_list_free(t1->vpd_list);
		free(t1);
	}
}

void
vpd_list_append(vpd_list_t ** vl,
		vpd_list_t *  new)
{
	assert((new == new->prev) && (new == new->next));
	
	if (NULL == *vl)
		*vl = new;
	else {
		new->next = *vl;
		new->prev = (*vl)->prev;

		(*vl)->prev->next = new;
		(*vl)->prev = new;
	}
}

void
vpd_list_prepend(vpd_list_t ** vl,
		 vpd_list_t *  new)
{
	vpd_list_append(vl, new);
	*vl = new;
}

static void
vpd_list_delete(vpd_list_t ** vpd, vpd_list_t * x)
{
	if (*vpd == (*vpd)->next)
		*vpd = NULL;
	else {
		x->prev->next = x->next;
		x->next->prev = x->prev;

		if (*vpd == x)
			*vpd = x->next;
			
		x->next = x;
		x->prev = x;
	}
}

bool
vpd_list_append_key_value(vpd_list_t **   vl,
			  const vpd_key_t k,
			  const unsigned char * v,
			  size_t          vlen)
{
	bool ret;
	vpd_list_t * t;

	ret = false;

	t = vpd_list_new_internal(k, v, vlen);
	if (NULL != t) {
		vpd_list_append(vl, t);
		ret = true;
	}

	return ret;
}

bool
vpd_list_prepend_key_value(vpd_list_t **   vl,
			   const vpd_key_t k,
			   const unsigned char * v,
			   size_t          vlen)
{
	bool ret;
	vpd_list_t * t;

	ret = false;

	t = vpd_list_new_internal(k, v, vlen);
	if (NULL != t) {
		vpd_list_prepend(vl, t);
		ret = true;
	}

	return ret;
}

bool
vpd_list_maybe_append_key_string(vpd_list_t **   vpd,
				 const char *    k,
				 const char *    v)
{
	bool ret;
	size_t vlen;

	ret = false;

	if ((NULL != v)) {
		vlen = strlen(v);
		if (0 != vlen)
			ret = vpd_list_append_key_value(vpd,
							(const unsigned char *) k,
							(unsigned char *) v,
							vlen);
	}

	return ret;
}

void
vpd_list_sort(vpd_list_t ** vpd,
	      const char *  order[])
{
	int i;
	vpd_list_t * out;
	vpd_list_t * x;
	vpd_list_t * t;

	out = vpd_list_new();

	/* This is inefficient, but the lists are usually short. */
	for (i = 0; NULL != order[i]; i++) {
		VPD_FOREACH_TEMP(*vpd, x, t) {
			if (0 == memcmp(x->key, order[i], sizeof(vpd_key_t))) {
				vpd_list_delete(vpd,  x);
				vpd_list_append(&out, x);
			}
		}
	}

	/* Sort remaining VPD. */
	while (NULL != *vpd) {
		x = NULL;
		VPD_FOREACH(*vpd, t) {
			if ((NULL == x) ||
			    (-1 == memcmp(t->key, x->key, sizeof(vpd_key_t))))
				x = t;
		}
		vpd_list_delete(vpd, x);
		vpd_list_append(&out, x);
	}

	*vpd = out;
}

vpd_list_list_t *
vpd_list_list_append(vpd_list_list_t * vll,
		     vpd_list_t *      vl)
{
	vpd_list_list_t * ret;

	ret = vpd_list_list_new_internal(vl);

	if (NULL != ret) {

		if (NULL == vll)
			vll = ret;
		else {
			ret->next = vll;
			ret->prev = vll->prev;
			
			vll->prev->next = ret;
			vll->prev = ret;
			
			ret = vll;
		}
	}

	return ret;
}

bool
vpd_list_to_text(const vpd_list_t * vl,
		 growbuf **         buf)
{
	bool ret;
	const vpd_list_t * t;

	ret = false;

	if ((NULL != *buf) && (NULL != vl)) {
		VPD_FOREACH(vl, t) {
			*buf = growbuf_append_byte(*buf, '*');
			*buf = growbuf_append(*buf, t->key, sizeof(t->key));
			*buf = growbuf_append_byte(*buf, ' ');
			*buf = growbuf_append(*buf,
					      (unsigned char *) t->val,
					      strlen(t->val));
			*buf = growbuf_append_byte(*buf, '\n');
		}
		ret = true;
	}

	return ret;
}

growbuf *
vpd_list_list_to_text(const vpd_list_list_t * vll)
{
	const vpd_list_list_t * t;

	growbuf * ret = growbuf_new(512);

	if (NULL != vll) {
		VPD_FOREACH(vll, t) {
			if (!vpd_list_to_text(t->vpd_list, &ret)) {
				growbuf_free(ret);
				return NULL;
			}
		}
	}

	return ret;
}

static char *
vpd_key_value_to_vpd_dir(const char *    vpd_dir,
			 const vpd_key_t k,
			 const char *    v)
{
	char * kk;
	char * f;
	short i;

	/*
	 * Could use an array for kk and snprintf, but then would have
	 * to strdup it at the end and that might fail after we've
	 * written the value to the file.
	 */
	f = NULL;
	kk = NULL;
	for (i = 0; i < 1000 ; i++) {
		if (0 == i) {
			if (-1 == asprintf(&kk, "%2.2s", k))
				kk = NULL;
		} else {
			if (-1 == asprintf(&kk, "%2.2s.%03d", k, i))
				kk = NULL;
		}

		if (NULL != kk) {
			f = lsvpd_relative_filename(vpd_dir, kk);
			if (NULL != f) {
				if (0 != access(f, F_OK))
					break;

				free(f);
				f = NULL;
			}
			
			free(kk);
			kk = NULL;
		}
	}

	if (NULL != f) {
		if (!lsvpd_file_write_string(f, v)) {
			free(kk);
			kk = NULL;
		}
		free(f);
	}

	return kk;
}

static void
vpd_dir_append_string_to_keys_file(FILE *       ksfd,
				   const char * s)
{
	fprintf(ksfd, "%s ", s);
}

bool
vpd_list_to_vpd_dir(const vpd_list_t * vl,
		    const char *       vpd_dir)
{
	bool ret;
	const vpd_list_t * t;
	char * s;
	FILE * ksfd;
	char * kk;

	ret = true;

	if (!lsvpd_mkdir(vpd_dir))
		return false;

	s = lsvpd_relative_filename(vpd_dir, VPD_LIST_FILE);		
	if (NULL == s)
		return false;
	ksfd = fopen(s, "w");
	free(s);
	if (NULL == ksfd)
		return false;

	VPD_FOREACH(vl, t) {
		kk = vpd_key_value_to_vpd_dir(vpd_dir, t->key, t->val);
		if (NULL != kk) {
			vpd_dir_append_string_to_keys_file(ksfd, kk);
			free(kk);
		} else
			ret = false;
	}

	fclose(ksfd);

	return ret;
}

bool
vpd_list_list_to_node(const vpd_list_list_t * vll,
		      const char *            node)
{
	const vpd_list_list_t * t;

	short subnode = 0;
	char * l;
	char * d;

	bool ret = true;
	
	if (NULL == (l = lsvpd_relative_filename(node, VPD_SUBDIR)))
		return false;

	if (!lsvpd_mkdir(l)) {
		free(l);
		return false;
	}
		
	VPD_FOREACH(vll, t) {
		if (-1 != asprintf(&d, "%s/%03d", l, subnode)) {
			ret = ret && vpd_list_to_vpd_dir(t->vpd_list, d);
			free(d);
			subnode++;
		} else
			ret = false;
	}

	free(l);

	return ret;
}

vpd_list_t *
vpd_text_to_vpd_list(const growbuf * buf)
{
	unsigned char * t1;
	unsigned char * t2;
	unsigned char * t3;

	vpd_list_t * vl;
	growbuf_iter * i;

	vl = vpd_list_new();

	i = growbuf_iter_new(buf);
	if (NULL == i) {
		vpd_list_free(vl);
		return NULL;
	}

	while (growbuf_iter_have(i, 4)) {
		growbuf_iter_next(i);
		t1 = growbuf_iter_current(i);
		growbuf_iter_move(i, 2);
		t2 = growbuf_iter_current(i);
		growbuf_iter_next(i);
		if (!growbuf_iter_more(i))
			break;

		t2 = growbuf_iter_current(i);
		if (NULL == (t3 = memchr(t2, '\n', growbuf_iter_remaining(i))))
			break;

		vpd_list_append_key_value(&vl, t1, t2, (size_t) (t3 - t2));
		growbuf_iter_move(i, (off_t) (t3 - t2) +1);
	}

	growbuf_iter_free(i);

	return vl;
}

static vpd_list_list_t *
vpd_list_to_vpd_list_list(const vpd_list_t * vl,
			  const vpd_key_t    sep)
{
	const vpd_list_t * t;

	vpd_list_t * c = NULL;

	vpd_list_list_t * ret = vpd_list_list_new();

	if (NULL != vl) {
		VPD_FOREACH(vl, t) {
			if (0 == memcmp(sep, t->key, sizeof(vpd_key_t))) {
				if (NULL != c)
					ret = vpd_list_list_append(ret, c);
				c = vpd_list_new();
			}
			
			 vpd_list_append_key_value(&c,
						   t->key,
						   (unsigned char *) t->val,
						   strlen(t->val));

			if (NULL == c) {
				vpd_list_list_free(ret);
				break;	
			}
		}
	}

	if ((NULL != c) && (NULL != ret))
		ret = vpd_list_list_append(ret, c);

	return ret;
}

vpd_list_list_t *
vpd_text_to_vpd_list_list(const growbuf * buf,
			  const vpd_key_t sep)
{
	/* There are marginally more efficient ways of doing this, but
	 * this is by far the simplest... */

	vpd_list_t * t = vpd_text_to_vpd_list(buf);

	vpd_list_list_t * ret = vpd_list_to_vpd_list_list(t, sep);

	vpd_list_free(t);

	return ret;
}

char *
vpd_dir_get_subdirnum(const char * dir,
		      int subdirnum)
{
	char * ret;

	ret = NULL;

	if ((NULL != dir) &&
	    (-1 == asprintf(&ret, "%s/%s/%03d", dir, VPD_SUBDIR, subdirnum)))
		ret = NULL;
	return ret;
}

char *
vpd_dir_get_default(const char * dir)
{
	return vpd_dir_get_subdirnum(dir, 0);
}

static void
vpd_dir_field_add_to_keys_file(const char *    vpd_dir,
			       const vpd_key_t key)
{
	char * s;
	char * out;
	char * t;


	s = lsvpd_file_relative_read_string(vpd_dir, VPD_LIST_FILE);

	// FIXME: Cheap and nasty, but simplifies code below.
	if (NULL == s)
		s = strdup("");

	out = NULL;
	
	if (0 == memcmp(key, "AX", sizeof(vpd_key_t)) &&
	    (NULL != (t = strstr(s, "DS"))) &&
	    (('\0' == t[2]) || (' ' == t[2]))) {
		/* Adding AX, try to put after DS. */
		memcpy(t, "%s", 2);
		if (-1 == asprintf(&out, s, "DS AX"))
			out = NULL;
	} else if (0 != memcmp(key, "YL", sizeof(vpd_key_t)) &&
		   (NULL != (t = strstr(s, "YL"))) &&
		   (('\0' == t[2]) || (' ' == t[2]))) {
			/* Not adding YL, try to put before YL. */
		memcpy(t, "%s", 2);
		if (-1 != asprintf(&t, "%2.2s YL", key)) {
			if (-1 == asprintf(&out, s, t))
				out = NULL;		
			free(t);
		}
	}
	
	if (NULL == out)
		if (-1 == asprintf(&out, "%s%2.2s ", s, key))
			out = NULL;
	
	if (NULL != out) {
		lsvpd_file_relative_write_string(vpd_dir, VPD_LIST_FILE, out);
		free(out);
	}

	free(s);
}

extern bool
vpd_dir_field_ensure(const char *    vpd_dir,
		     const vpd_key_t key,
		     const char *    value)
{
	bool ret;
	char * f;

	ret = false;

	if (-1 != asprintf(&f, "%s/%2.2s", vpd_dir, key)) {
		if (0 != access(f, F_OK))
			if (lsvpd_file_write_string(f, value)) {
				vpd_dir_field_add_to_keys_file(vpd_dir, key);
				ret = true;
			}
		free(f);
	}

	return ret;
}
